from social.utils import module_member
from social.backends.base import BaseAuth
from social.apps.django_app.utils import setting


DEFAULT_AUTHENTICATION_PROVIDERS_BRIEF = 3


def get_login_providers(request, short=False):
    """
    Returns a list of available login providers based on the
    AUTHENTICATION_BACKENDS setting. Each provider is represented as a
    dictionary containing the backend name, name of required parameter if
    required and its verbose name.
    """
    def extract_backend_data(klass):
        """
        Helper function which extracts information useful for use in
        templates from SocialAuth subclasses and returns it as a
        dictionary.
        """
        return {
            'name': klass.name,
            'required_field': klass.REQUIRED_FIELD_NAME,
            'required_field_verbose': klass.REQUIRED_FIELD_VERBOSE_NAME,
        }

    backends = (module_member(auth_backend) for auth_backend in setting('AUTHENTICATION_BACKENDS'))
    providers = [extract_backend_data(backend) for backend in backends if issubclass(backend, BaseAuth)]
    if short:
        return providers[:setting('AUTHENTICATION_PROVIDERS_BRIEF',
                                  DEFAULT_AUTHENTICATION_PROVIDERS_BRIEF)]
    return providers


def login_providers(request):
    """
    Returns the full list of login providers as the social_auth context
    variable.
    """
    return {'login_providers': get_login_providers(request)}


def login_providers_short(request):
    """
    Returns the short list of login providers for use in a login widget as
    the social_auth context variable.
    """
    return {'login_providers_short': get_login_providers(request, short=True)}


def login_providers_both(request):
    """
    Returns both the short and the long list of login providers.
    """
    return {
        'login_providers': get_login_providers(request),
        'login_providers_short': get_login_providers(request, short=True),
    }



from django.contrib.auth.forms import (UserCreationForm,
    PasswordChangeForm as AuthPasswordChangeForm)
from django.contrib.auth.models import User
from django.forms.models import ModelForm
from django.utils.translation import string_concat, ugettext_lazy as _
from social.apps.django_app.utils import setting

from ksp_login import SOCIAL_AUTH_PARTIAL_PIPELINE_KEY
from ksp_login.utils import import_string


class KspUserCreationForm(UserCreationForm):
    """
    A custom user creation form that can make the password fields
    optional.
    """
    class Meta(UserCreationForm.Meta):
        fields = ('username', 'first_name', 'last_name', 'email')

    def __init__(self, *args, **kwargs):
        """
        If the user is registering via an external service, this gathers
        initial data from the service and marks the password fields as
        optional, otherwise it just sets the help_text of the first
        password field.
        """
        try:
            request = kwargs['request']
            del kwargs['request']
        except KeyError:
            raise TypeError("Argument 'request' missing.")
        try:
            pipeline_state = request.session[setting('SOCIAL_AUTH_PARTIAL_PIPELINE_KEY',
                                                     SOCIAL_AUTH_PARTIAL_PIPELINE_KEY)]
            pipeline_state = pipeline_state['kwargs']
            self.password_required = False
        except KeyError:
            self.password_required = True
            pipeline_state = None
        if not args and 'initial' not in kwargs:
            kwargs['initial'] = self.get_initial_from_pipeline(pipeline_state)

        super(KspUserCreationForm, self).__init__(*args, **kwargs)

        self.fields['password1'].help_text = _(
            "We recommend choosing a strong passphrase but we don't "
            "enforce any ridiculous constraints on your passwords."
        )

        if not self.password_required:
            self.fields['password1'].required = False
            self.fields['password2'].required = False
            self.fields['password1'].help_text = string_concat(_(
                "Since you're logging in using an external provider, "
                "this field is optional; however, by supplying it, you "
                "will be able to log in using a password. "
            ), self.fields['password1'].help_text)

    def get_initial_from_pipeline(self, pipeline_state):
        """
        Returns a dictionary with initial field values extracted from the
        social_auth pipeline state.
        """
        return None if not pipeline_state else {
            'username': pipeline_state['details']['username'],
            'first_name': pipeline_state['details']['first_name'],
            'last_name': pipeline_state['details']['last_name'],
            'email': pipeline_state['details']['email'],
        }

    def clean_password2(self):
        """
        We need to validate the submitted passwords if they are required
        or either has been submitted, otherwise the user account will be
        passwordless.
        """
        if (self.password_required or self.cleaned_data.get('password1') or
                self.cleaned_data.get('password2')):
            return super(KspUserCreationForm, self).clean_password2()
        return None

    def save(self, commit=True):
        """
        If a password was provided or is required, just delegate to the
        parent's save, otherwise explicitly set an unusable password.
        """
        user = super(KspUserCreationForm, self).save(commit=False)
        if not (self.cleaned_data.get('password1') or
                self.password_required):
            user.set_unusable_password()
        if commit:
            user.save()
        return user


class UserProfileForm(ModelForm):
    """
    A simple form that allows an existing user to change their e-mail and
    name.
    """
    class Meta:
        model = User
        fields = ('first_name', 'last_name', 'email')

    def __init__(self, *args, **kwargs):
        user = kwargs.pop('user')
        if user:
            kwargs['instance'] = user
        super(UserProfileForm, self).__init__(*args, **kwargs)


class PasswordChangeForm(AuthPasswordChangeForm):
    """
    A form that lets a user change or optionally remove their password
    after entering their current password.
    """
    def __init__(self, password_required=True, *args, **kwargs):
        super(PasswordChangeForm, self).__init__(*args, **kwargs)
        self.password_required = password_required

        if not password_required:
            self.fields['new_password1'].required = False
            self.fields['new_password2'].required = False
            self.fields['new_password1'].help_text = _(
                "Leave this field empty to disable password-based access "
                "to your account."
            )

    def clean_new_password2(self):
        """
        We need to validate the submitted passwords if they are required
        or either has been submitted, otherwise the user account will be
        passwordless.
        """
        if (self.password_required or self.cleaned_data.get('new_password1') or
                self.cleaned_data.get('new_password2')):
            return super(PasswordChangeForm, self).clean_new_password2()
        return None

    def save(self, commit=True):
        """
        If a password was provided or is required, just delegate to the
        parent's save, otherwise explicitly set an unusable password.
        """
        user = super(PasswordChangeForm, self).save(commit=False)
        if not (self.cleaned_data.get('new_password1') or
                self.password_required):
            user.set_unusable_password()
        if commit:
            user.save()
        return user


class BaseUserProfileForm(ModelForm):
    """
    Base class for any additional user info forms. This implements the
    contract required by the register and settings views.
    """

    # Override this attribute if the ForeignKey to User in your model is
    # called something other than 'user'.
    USER_MODEL_FIELD = 'user'

    def __init__(self, *args, **kwargs):
        kwargs.pop('request', None)
        user = kwargs.pop('user', None)
        self.user = user
        if user is not None:
            inst_kw = {self.USER_MODEL_FIELD: user}
            instance, created = self._meta.model.objects.get_or_create(**inst_kw)
            kwargs['instance'] = instance
        super(BaseUserProfileForm, self).__init__(*args, **kwargs)

    def set_user(self, user):
        self.user = user

    def save(self, commit=True):
        instance = super(BaseUserProfileForm, self).save(commit=False)
        if self.user is not None:
            setattr(instance, self.USER_MODEL_FIELD, self.user)
        if commit:
            instance.save()
        return instance


def get_profile_forms():
    form_names = setting('KSP_LOGIN_PROFILE_FORMS', [])
    return [import_string(name) for name in form_names]



try:
    from django.utils.module_loading import import_string
except ImportError:
    from importlib import import_module

    from django.utils import six

    def import_string(dotted_path):
        """
        Import a dotted module path and return the attribute/class designated by the
        last name in the path. Raise ImportError if the import failed.

        Borrowed from Django 1.7.
        """
        try:
            module_path, class_name = dotted_path.rsplit('.', 1)
        except ValueError:
            msg = "%s doesn't look like a module path" % dotted_path
            six.reraise(ImportError, ImportError(msg), sys.exc_info()[2])

        module = import_module(module_path)

        try:
            return getattr(module, class_name)
        except AttributeError:
            msg = 'Module "%s" does not define a "%s" attribute/class' % (
                dotted_path, class_name)
            six.reraise(ImportError, ImportError(msg), sys.exc_info()[2])



from os.path import join, dirname
from setuptools import setup


LONG_DESCRIPTION = """
KSP Login is an app for easy authentication management with support for
social as well as traditional authentication.
"""


def long_description():
    """Return long description from README.rst if it's present
    because it doesn't get installed."""
    try:
        return open(join(dirname(__file__), 'README.rst')).read()
    except IOError:
        return LONG_DESCRIPTION


install_requires = [
    'Django>=1.6',
    'python-social-auth',
]


try:
    import importlib
except ImportError:
    install_requires.append('importlib')


setup(
    name='ksp-login',
    version='0.3.2',
    author='Michal Petrucha',
    author_email='michal.petrucha@koniiiik.org',
    url='https://github.com/koniiiik/ksp_login',
    packages=['ksp_login', 'ksp_login.templatetags'],
    package_data={
        'ksp_login': [
            'locale/*/*/*.[mp]o',
            'static/ksp_login/css/*.css',
            'static/ksp_login/img/*.svg',
            'static/ksp_login/img/*.png',
            'static/ksp_login/img/*.txt',
            'static/ksp_login/js/*.js',
            'templates/ksp_login/*.html',
            'templates/ksp_login/parts/*.html',
        ],
    },
    license='LICENSE',
    description='A Django app for both traditional and social authentication',
    long_description=long_description(),
    install_requires=install_requires,
)



from django.shortcuts import render
from django.core.urlresolvers import reverse
from django.shortcuts import redirect
from django.utils.translation import ugettext_lazy as _
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.contrib.auth.views import (login as auth_login, logout as
    auth_logout, password_change)
from django.contrib.auth.forms import SetPasswordForm
from social.apps.django_app.default.models import UserSocialAuth
from social.apps.django_app.utils import setting
from social.apps.django_app.views import disconnect as social_disconnect
from ksp_login import SOCIAL_AUTH_PARTIAL_PIPELINE_KEY
from ksp_login.context_processors import login_providers
from ksp_login.forms import (KspUserCreationForm, PasswordChangeForm,
    UserProfileForm, get_profile_forms)


def login(request):
    if request.user.is_authenticated():
        next_page = request.REQUEST.get('next', 'account_settings')
        return redirect(next_page)
    return auth_login(request, template_name='ksp_login/login.html',
                      extra_context=login_providers(request))


def logout(request):
    response = auth_logout(request, next_page='/')
    messages.success(request, _("Logout successful"))
    return response


@login_required
def disconnect(request, backend, association_id):
    """
    If the user has at least one other social account association or a
    valid password, disconnects the given social account, otherwise asks
    the user to set up a password before proceeding.
    """
    associations = UserSocialAuth.get_social_auth_for_user(request.user)
    has_assoc = associations.exclude(id=association_id).count()
    has_pass = request.user.has_usable_password()
    if has_assoc or has_pass:
        return social_disconnect(request=request, backend=backend,
                                 association_id=association_id)
    return render(request, 'ksp_login/invalid_disconnect.html')


@login_required
def password(request):
    """
    Sets, changes or removes the currently logged in user's passwords,
    depending on whether they have any social account associations.
    """
    has_assoc = UserSocialAuth.get_social_auth_for_user(request.user).count()
    if request.user.has_usable_password():
        def form(*args, **kwargs):
            return PasswordChangeForm(not has_assoc, *args, **kwargs)
    else:
        form = SetPasswordForm
    return password_change(request,
                           post_change_redirect=reverse('account_settings'),
                           password_change_form=form,
                           template_name='ksp_login/password.html')


def register(request, creation_form=KspUserCreationForm):
    """
    As the name suggests, registers a new user.

    Can replace the create_user function in the SOCIAL_AUTH pipeline
    (through ksp_login.pipeline.register_user, with the corresponding
    save_status_to_session, of course) to make it possible for the user to
    pick a username.

    Also presents additional app-specific forms listed in the
    KSP_LOGIN_PROFILE_FORMS setting to the user.
    """
    if request.user.is_authenticated():
        return redirect('account_settings')
    try:
        pipeline_state = request.session[setting('SOCIAL_AUTH_PARTIAL_PIPELINE_KEY',
                                                 SOCIAL_AUTH_PARTIAL_PIPELINE_KEY)]
        backend = pipeline_state['backend']
        pipeline_state = pipeline_state['kwargs']
        standalone = False
    except KeyError:
        standalone = True

    form_classes = [creation_form] + get_profile_forms()

    if request.method == "POST":
        forms = [form(request.POST, request=request)
                 for form in form_classes]
        if all(form.is_valid() for form in forms):
            user = forms[0].save()
            for form in forms[1:]:
                form.set_user(user)
                form.save()
            if standalone:
                return redirect('account_login')
            pipeline_state['user'] = user.id
            # This ensures the top-level session dict changes, otherwise
            # our changes in pipeline_state might not get stored.
            request.session.setdefault('ksp_login_dummy_key', 0)
            request.session['ksp_login_dummy_key'] += 1
            return redirect('social:complete', backend=backend)
    else:
        forms = [form(request=request) for form in form_classes]

    return render(request, "ksp_login/registration.html", {
        'forms': forms,
    })


@login_required
def settings(request, settings_form=UserProfileForm):
    """
    Presents the user a form with their settings, basically the register
    form minus username minus password.

    Also presents additional app-specific forms listed in the
    KSP_LOGIN_PROFILE_FORMS setting to the user.
    """
    form_classes = [settings_form] + get_profile_forms()

    if request.method == "POST":
        forms = [form(request.POST, user=request.user)
                 for form in form_classes]
        if all(form.is_valid() for form in forms):
            for form in forms:
                form.save()
            return redirect('account_settings')
    else:
        forms = [form(user=request.user) for form in form_classes]

    return render(request, 'ksp_login/settings.html', {
        'account_associations': UserSocialAuth.get_social_auth_for_user(request.user),
        'forms': forms,
    })



from django.conf.urls import patterns, include, url
from django.core.urlresolvers import reverse_lazy
from django.views.generic.base import RedirectView

urlpatterns = patterns('ksp_login.views',
    url(r'^login/$', 'login', name='account_login'),
    url(r'^$', 'settings', name='account_settings'),
    url(r'^logout/$', 'logout', name='account_logout'),
    url(r'^register/$', 'register', name='account_register'),
    url(r'^disconnect/(?P<backend>[^/]+)/(?P<association_id>[^/]+)/$',
        'disconnect', name='account_disconnect'),
    url(r'^password/$', 'password', name='account_password'),
)

urlpatterns += patterns('',
    url(r'', include('social.apps.django_app.urls', namespace='social')),
    url(r'^settings/', RedirectView.as_view(url=reverse_lazy('account_settings'))),
)
